/*
 *  Geometry.h
 *  Painter3D
 *
 *  Created by Cindy Grimm on 2/27/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef _GEOMETRY_DEFS_H_
#define _GEOMETRY_DEFS_H_

#include "SystemSpecific.h"

/** 
 \class R2PointTC Rn_Defs.H utils/Rn_Defs.H
 \ingroup SimpleGeometryElems
 \brief A 2D Point
 
 Two numbers.  Type-safe addition/subtraction/multiplication for points,
 vectors, and co-vectors. See operators in documentation for
 non-member functions.
 
 */
template<class Coord>
class R2PointTC {
    public:
        /**@name Constructors */
        //@{
        ///
        R2PointTC() { }
        /// Initialize with given values
        R2PointTC(const Coord u, const Coord v);
        //@}
        
        /**@name Access grants */
        //@{
        /// Return dimension (always 2)
        int Dim() const { return 2; }
        
        /// Does a check that int is in bounds, i.e., 0,1
        Coord& operator[] (  const int i ) { ASSERT(i == 0 || i == 1); return (&u)[i]; }
        /// Does a check that int is in bounds, i.e., 0,1
        const Coord& operator[] (  const int i ) const { ASSERT(i == 0 || i == 1); return (&u)[i]; }
        //@}
        
    private:
        ///
        Coord u, v;
};

template<class Coord>
inline
double Length( const R2PointTC<Coord> &in_pt ) 
{
    return sqrt( pow(in_pt[0], 2) + pow(in_pt[1], 2) );
}

template<class Coord>
inline
double Length( const R2PointTC<Coord> &in_pt1, const R2PointTC<Coord> &in_pt2 ) 
{
    return sqrt( pow(in_pt1[0] - in_pt2[0], 2) + pow(in_pt1[1] - in_pt2[1], 2) );
}

template<class Coord>
inline
R2PointTC<Coord>::R2PointTC ( const Coord _u, const Coord _v )
{
    u = _u; v = _v;
}

///
template<class Coord>
inline ostream& operator<<( ostream &out, const R2PointTC<Coord> &p )
{
    out << p[0] << " " << p[1] << " ";
    return out;
}

///
template<class Coord>
inline istream& operator>>( istream &in, R2PointTC<Coord> &p )
{
    in >> p[0] >> p[1];
    return in;
}



/** 
 \class R3PointTC Rn_Defs.H utils/Rn_Defs.H
 \ingroup SimpleGeometryElems
 \brief A 3D point.
 
 Three numbers.  Type-safe addition/subtraction/multiplication for points,
 vectors, and co-vectors. See operators in documentation for
 non-member functions.
 
 */
template<class Coord>
class R3PointTC
    {
    public:
        /// Default constructor
        R3PointTC() { }
        /// Initialize with given values
        inline R3PointTC( const Coord& in_x, const Coord& in_y, const Coord& in_z ) { x = in_x; y = in_y; z = in_z; }
        
        /// Return dimension (always 3)
        int Dim() const { return 3; }
        
        /**@name Access grants */
        //@{
        /// Does a check that int is in bounds, i.e., 0,1,2
        Coord& operator[] ( const int i ) { ASSERT(i >= 0 && i <= 2 ); return (&x)[i]; }
        /// Does a check that int is in bounds, i.e., 0,1,2
        const Coord& operator[] ( const int i ) const { ASSERT(i >= 0 && i <= 2 ); return (&x)[i]; }
        //@}
    private:
        ///
        Coord x, y, z;
};

/** 
 \class R3VectorTC Rn_Defs.H utils/Rn_Defs.H
 \ingroup SimpleGeometryElems
 \brief A 3D Column vector.
 
 Three numbers.  Type-safe addition/subtraction/multiplication for points,
 vectors, and co-vectors. See operators in documentation for
 non-member functions.
 
 */
template<class Coord>
class R3VectorTC
    {
    public:
        ///
        R3VectorTC() {}
        /// Initialize with given values
        inline R3VectorTC( const Coord& in_dx, const Coord& in_dy, const Coord& in_dz ) { dx = in_dx; dy = in_dy; dz = in_dz; }
        
        /// Return dimension (always 3)
        int Dim() const { return 3; }
        
        /**@name Access grants */
        //@{
        /// Does a check that int is in bounds, i.e., 0,1,2
        Coord& operator[] ( const int i ) { ASSERT(i >= 0 && i <= 2 ); return (&dx)[i]; }
        /// Does a check that int is in bounds, i.e., 0,1,2
        const Coord& operator[] ( const int i ) const { ASSERT(i >= 0 && i <= 2 ); return (&dx)[i]; }
        //@}
        
        /**@name Assignment operators */
        //@{
        /// 
        inline R3VectorTC<Coord>& operator += ( const R3VectorTC<Coord>& );
        ///
        inline R3VectorTC<Coord>& operator -= ( const R3VectorTC<Coord>& );
        ///
        inline R3VectorTC<Coord>& operator *= ( const Coord& );
        ///
        inline R3VectorTC<Coord>& operator /= ( const Coord& );
        //@}
        
        /**@name Self-editing */
        //@{
        /// Divide by Length; sets length to 1.0 if not zero. Fails if length is zero.
        inline Coord Normalize() {
            const Coord dLen = sqrt(dx * dx + dy * dy + dz * dz);
            if ( fabs( dLen ) > 1e-32 ) {
                dx = dx / dLen;
                dy = dy / dLen;
                dz = dz / dLen;
            }
            
            return dLen;
        }
        
        //@}
        
        /**@name Unary operators */
        //@{
        /// Return a vector that is perpendicular to this one (not unique)
        inline R3VectorTC<Coord>  Perpendicular() const;
        //@}
        
    private:
        ///
        Coord dx, dy, dz;
};

/// Dot product of two vectors
template<class Coord>
inline Coord
Dot( const R3VectorTC<Coord>& v1, const R3VectorTC<Coord>& v2 )
{
    return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2];
}


/** 
 \class R4MatrixTC Rn_Defs.H utils/Rn_Defs.H
 \ingroup SimpleGeometryElems
 \brief A 4x4 matrix
 
 Four numbers. Data must be in column major order to match OpenGL Type-safe addition/subtraction/multiplication for points,
 vectors, and co-vectors. See operators in documentation for
 non-member functions.
 
 */
template< class Coord >
class R4MatrixTC {
    public:
        ///
        R4MatrixTC() {}
        
        /**@name Access grants */
        //@{
        ///
        inline Coord & operator () ( const int iRow, const int iCol ) { return m_mat[iCol][iRow]; }
        ///
        inline const Coord & operator () ( const int iRow, const int iCol ) const { return m_mat[iCol][iRow]; }
        //@}
        
        /**@name Assignment operators */
        //@{
        /// 
        inline R4MatrixTC<Coord>& operator *= ( const R4MatrixTC<Coord>& );
        ///
        inline R4MatrixTC<Coord>& operator += ( const R4MatrixTC<Coord>& );
        ///
        inline R4MatrixTC<Coord>& operator -= ( const R4MatrixTC<Coord>& );
        ///
        inline R4MatrixTC<Coord>& operator *= ( const Coord );
        ///
        inline R4MatrixTC<Coord>& operator /= ( const Coord );
        //@}
        
        /**@name Unary operators */
        //@{
        /// 
        inline R4MatrixTC<Coord> operator + ( );
        ///
        inline R4MatrixTC<Coord> operator - ( );
        //@}
        
        /**@name Matrix construction operators */
        //@{
        /// Set self to identity
        void MakeIdentity( );
        /// Static method that returns an identity matrix
        static R4MatrixTC<Coord> Identity( );
        /** \brief Static method that returns a rotation matrix 
         @param axis is one of 0,1,2 (x,y,z)
         @param angle Angle in Radians
         @returns matrix */
        static R4MatrixTC<Coord> Rotation(const int axis, const Coord angle);
        /// Static method that returns a scale matrix
        static R4MatrixTC<Coord> Scaling( const Coord in_x, const Coord in_y, const Coord in_z, const Coord in_w );
        /// Static method that returns a uniform scaling matrix
        static R4MatrixTC<Coord> Scaling( const Coord in_scl );
        /// Static method that returns a translation matrix for a 3D point/vector
        static R4MatrixTC<Coord> Translation( const Coord in_x, const Coord in_y, const Coord in_z );
        //@}
        
        /**@name Matrix operations */
        //@{
        /** \brief Take the inverse of the matrix
         @param out_bSuccess If the determinant was non-zero, i.e., inverse exists
         @returns matrix inverse */
        inline R4MatrixTC<Coord> Inverse( bool &out_bSuccess ) const;
        /// 
        inline R4MatrixTC<Coord> Transpose(  ) const;
        ///
        inline Coord Determinant () const;
        //@}
                
    private:
        /// column row
        Coord m_mat[4][4];
};

///
template<class Coord>
inline R4MatrixTC<Coord> R4MatrixTC<Coord>::Scaling( const Coord in_x, const Coord in_y, const Coord in_z, const Coord in_w )
{
    R4MatrixTC<Coord> out_mat;
    out_mat.MakeIdentity(  );
    
    out_mat(0,0) = in_x;
    out_mat(1,1) = in_y;
    out_mat(2,2) = in_z;
    out_mat(3,3) = in_w;
    
    return out_mat;
}

template<class Coord>
inline R4MatrixTC<Coord> R4MatrixTC<Coord>::Translation( const Coord in_d1, const Coord in_d2, const Coord in_d3 )
{
    R4MatrixTC<Coord> out_mat;
    out_mat.MakeIdentity( );
    
    out_mat(0,3) = in_d1;
    out_mat(1,3) = in_d2;
    out_mat(2,3) = in_d3;
    
    return out_mat;
}


template<class Coord>
inline void R4MatrixTC<Coord>::MakeIdentity(  )
{
    for (int i = 0; i < 4; i++) {
        for ( int j = 0; j < 4; j++) {
            if ( i == j )
                (*this)(i,j) = 1;
            else
                (*this)(i,j) = 0;
        }
    }
}

/// Multiply a vector by a scaler
template<class Coord>
inline R3VectorTC<Coord>
operator * ( const R3VectorTC<Coord>& v, const Coord& s )
{
    return R3VectorTC<Coord>( v[0] * s, v[1] * s, v[2] * s );
}


/// Add two vectors
template<class Coord>
inline R3VectorTC<Coord>
operator + ( const R3VectorTC<Coord>& v1, const R3VectorTC<Coord>& v2 )
{
    return R3VectorTC<Coord>( v1[0] + v2[0], v1[1] + v2[1], v1[2] + v2[2] );
}


/// Subtract two vectors
template<class Coord>
inline R3VectorTC<Coord>
operator - ( const R3VectorTC<Coord>& v1, const R3VectorTC<Coord>& v2 )
{
    return R3VectorTC<Coord>( v1[0] - v2[0], v1[1] - v2[1], v1[2] - v2[2] );
}

/// Returns non-negative number
template<class Coord>
inline Coord
Length( const R3VectorTC<Coord>& v )
{
    return (Coord) sqrt( v[0]*v[0] + v[1]*v[1] + v[2]*v[2] );
}

/// Return unit length vector, or vector if length is small (1e-12)
template<class Coord>
inline R3VectorTC<Coord>
UnitSafe( const R3VectorTC<Coord>& v )
{
    const Coord dLength = Length( v );
    
    if ( fabs( dLength ) < (Coord) 1e-12 )
        return v;
    
    const Coord dDiv = 1.0 / dLength;
    return v * dDiv;
}


/// Cross product
template<class Coord>
inline R3VectorTC<Coord>
Cross( const R3VectorTC<Coord>& v1, const R3VectorTC<Coord>& v2 )
{
    return R3VectorTC<Coord>(v1[1]*v2[2] - v1[2]*v2[1],
                             v1[2]*v2[0] - v1[0]*v2[2],
                             v1[0]*v2[1] - v1[1]*v2[0]);
}


/// Subtract two points
template<class Coord>
inline R3VectorTC<Coord>
operator - ( const R3PointTC<Coord>& p1, const R3PointTC<Coord>& p2 )
{
    return R3VectorTC<Coord>( p1[0] - p2[0], p1[1] - p2[1], p1[2] - p2[2] );
}


/// Multiply a matrix by a matrix
template<class Coord>
inline R4MatrixTC<Coord> 
operator * ( const R4MatrixTC<Coord>& m1, const R4MatrixTC<Coord>& m2 )
{
    R4MatrixTC<Coord> ret;
    
    for (int i = 0; i < 4; i += 1)
        for (int j = 0; j < 4; j += 1)
            ret(i, j) = m1(i, 0) * m2(0,j) + m1(i,1) * m2(1, j) + m1(i,2) * m2(2,j) + m1(i,3) * m2(3,j);
    return ret;
}


/** \brief Are the two numbers within epsilon of each other? Double precision
 @param in_d1 First number
 @param in_d2 Second number
 @param in_dEpsilon default is RNEpsilon_d
 */
inline bool RNApproxEqual( const double in_d1, const double in_d2,
                           const double in_dEpsilon = 1e-32)
{
    if (fabs(in_d1 - in_d2) > in_dEpsilon) return false;
    return true;
}


/** \ingroup SimpleGeometryElems Typedefs */
//@{
/// From R2PointTC
typedef R2PointTC<double> R2Pt;
/// From R2PointTC
typedef R2PointTC<float>  R2Pt_f;
/// From R2PointTC
typedef R2PointTC<int>    R2Pt_i;

/// From R3PointTC
typedef R3PointTC<double> R3Pt;
/// From R3PointTC
typedef R3PointTC<float>  R3Pt_f;
/// From R3PointTC
typedef R3PointTC<int>    R3Pt_i;

/// From R3VectorTC
typedef R3VectorTC<double> R3Vec;
/// From R3VectorTC
typedef R3VectorTC<float>  R3Vec_f;
/// From R3VectorTC
typedef R3VectorTC<int>    R3Vec_i;

/// From R4MatrixTC
typedef R4MatrixTC<double> R4Matrix;

//@}


#endif
