/*
 *  Images.h
 *  Painter3D
 *
 *  Created by Cindy Grimm on 2/27/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef __RAW_IMAGES_DEFS_H
#define __RAW_IMAGES_DEFS_H

#include "SystemSpecific.h"
#include "Geometry.h"
#include "UTILSColor.h"

/** Image manipulation.
 
 An image. See include/OpenGL/OGL_Images for OpenGL calls. Reads/writes bmps.
 
 <br>Files:
 <dd><li> Image.H
 <dd><li> Image.cpp
 */
class IMAGERaw {
protected:
    typedef enum {
        RED = 0, 
        GREEN = 1,
        BLUE = 2,
        ALPHA = 3
    } MyRGBOffset;
    
    int            m_iXSize, m_iYSize, m_iStep;
    bool           m_bHasAlpha;
    vector<int>    m_aiXIndx;
    vector<float>  m_afImage;
    
public:
    /**@name Accessors */
    //@{
    ///
    bool    HasAlpha() const { return m_bHasAlpha; }
    ///
    int     Width()  const { return m_iXSize; }
    ///
    int     Height() const { return m_iYSize; }
    ///
    int     NPixels() const { return m_iXSize * m_iYSize; }
    
    ///
    float     operator()( const int in_iX, const int in_iY, const int in_iRGB ) const;
    ///
    UTILSColor operator()( const R2Pt & ) const;
    ///
    UTILSColor operator()( const int in_iX, const int in_iY ) const;
    ///
    UTILSColor operator()( const int in_iPixel ) const;
    ///
    int        Index     ( const int in_iX, const int in_iY ) const;
    ///
    int        Index     ( const int in_iPixel ) const { return in_iPixel * m_iStep; }
    ///
    int        X         ( const int in_iIndex ) const;
    ///
    int        Y         ( const int in_iIndex ) const;
    ///
    float     Alpha( const int in_iX, const int in_iY ) const;
    ///
    float     Alpha( const int in_iPixel ) const;
    ///
    UTILSColor Interpolate( const R2Pt &in_pt ) const;
    ///
    vector<R2Pt> InterpolateIndex( const R2Pt &in_pt ) const;
    ///
    const float * Data() const { return &m_afImage[0]; }
    ///
    float * Data() { return &m_afImage[0]; }
    
    ///
    bool    IndexOk( const int in_iX, const int in_iY ) const;
    /// Does min/max
    UTILSColor Safe( const int in_iX, const int in_iY ) const;
    /// Does min/max
    float Safe( const int in_iX, const int in_iY, const int in_iRGB ) const;
    
    //@}
    
    /**@name Assignements */
    //@{
    ///
    void   SetSize( const int in_iX, const int in_iY, bool in_bHasAlpha = false );
    
    ///
    float &operator()( const int in_iX, const int in_iY, const int in_iRGB ) ;
    ///
    void    Set       ( const int in_iPixel, const UTILSColor &in_pix );
    ///
    void    Set       ( const int in_iX, const int in_iY, const UTILSColor &in_pix );
    ///
    void    Set       ( const UTILSColor &in_pix );
    ///
    void    SetAlpha  ( const int in_iX, const int in_iY, const float in_dAlpha );
    ///
    void    SetAlpha  ( const int in_iPixel, const float in_dAlpha );
    ///
    void    SetAlpha  ( const float in_dAlpha );
    
    //@}
    
    /**@name Resampling */
    //@{
    ///
    IMAGERaw &operator=( const IMAGERaw & ) ;
    //@}
    
    ///
    IMAGERaw();
    ///
    IMAGERaw( const int in_iX, const int in_iY, const bool in_bAlpha = false );
    ///
    ~IMAGERaw();
    
    /**@name Read/write */
    //@{
    /// Format determined by filename extension
    bool Read( const char *in_strFName );
    /// Format determined by filename extension
    bool Write( const char *in_strFName, const bool in_bUseAlpha = false) const;
    //@}
};

inline int IMAGERaw::Index( const int in_iX, const int in_iY ) const
{
    ASSERT( in_iX >=0 && in_iX < Width() );
    return m_aiXIndx[ in_iY ] + m_iStep * in_iX;
}

inline float  IMAGERaw::operator()( const int in_iX, const int in_iY, const int in_iRGB ) const
{
    return m_afImage[ in_iRGB + Index(in_iX, in_iY) ];
}

inline float &IMAGERaw::operator()( const int in_iX, const int in_iY, const int in_iRGB ) 
{
    return m_afImage[ in_iRGB + Index(in_iX, in_iY) ];
}

inline UTILSColor IMAGERaw::operator()( const R2Pt &in_pt ) const
{
    const int iIndx =  Index( WINminmax( (int) (in_pt[0] * m_iXSize + 0.5), 0, m_iXSize - 1 ),
                              WINminmax( (int) (in_pt[1] * m_iYSize + 0.5), 0, m_iYSize - 1 ) );
    
    return UTILSColor( m_afImage[ RED + iIndx ],
                      m_afImage[ GREEN + iIndx ],
                      m_afImage[ BLUE + iIndx ] );
}

inline UTILSColor IMAGERaw::operator()( const int in_iX, const int in_iY ) const
{
    const int iIndx = Index(in_iX, in_iY);
    
    return UTILSColor( m_afImage[ RED + iIndx ],
                      m_afImage[ GREEN + iIndx ],
                      m_afImage[ BLUE + iIndx ] );
}

inline UTILSColor IMAGERaw::operator()( const int in_iIndx ) const
{
    const int iIndx = Index( in_iIndx );
    return UTILSColor( m_afImage[ RED + iIndx ],
                      m_afImage[ GREEN + iIndx ],
                      m_afImage[ BLUE + iIndx ] );
}

inline float IMAGERaw::Alpha( const int in_iX, const int in_iY ) const
{
    const int iIndx = Index(in_iX, in_iY);
    
    return m_afImage[ ALPHA + iIndx ];
}

inline float IMAGERaw::Alpha( const int in_iIndex ) const
{
    const int iIndx = Index( in_iIndex );
    return m_afImage[ ALPHA + iIndx ];
}

inline void IMAGERaw::Set( const int in_iX, const int in_iY, const UTILSColor &in_pix ) 
{
    const int iIndx = Index(in_iX, in_iY);
    
    m_afImage[ RED + iIndx ] = (float) in_pix[0];
    m_afImage[ GREEN + iIndx ] = (float) in_pix[1];
    m_afImage[ BLUE + iIndx ] = (float) in_pix[2];
}

inline void IMAGERaw::Set( const int in_iIndx, const UTILSColor &in_pix ) 
{
    const int iIndx = Index( in_iIndx );
    m_afImage[ RED + iIndx ] = (float) in_pix[0];
    m_afImage[ GREEN + iIndx ] = (float) in_pix[1];
    m_afImage[ BLUE + iIndx ] = (float) in_pix[2];
}

inline void IMAGERaw::SetAlpha( const int in_iX, const int in_iY, const float in_dAlpha ) 
{
    const int iIndx = Index(in_iX, in_iY);
    
    ASSERT( m_bHasAlpha == true );
    m_afImage[ ALPHA + iIndx ] = in_dAlpha;
}

inline void IMAGERaw::SetAlpha( const int in_iIndx, const float in_dAlpha ) 
{
    ASSERT( m_bHasAlpha == true );
    const int iIndx = Index( in_iIndx );
    m_afImage[ ALPHA + iIndx ] = in_dAlpha;
}

inline int IMAGERaw::X( const int in_iIndex ) const
{
    return (in_iIndex / m_iStep) % Width();
}

inline int IMAGERaw::Y( const int in_iIndex ) const
{
    return (in_iIndex / m_iStep) / Width();
}

inline bool IMAGERaw::IndexOk( const int in_iX, const int in_iY ) const
{
    if ( in_iX < 0 || in_iY < 0 )
        return false;
    
    if ( in_iX >= Width() || in_iY >= Height() )
        return false;
    
    return true;
}

inline UTILSColor IMAGERaw::Safe( const int in_iX, const int in_iY ) const
{
    return (*this)( WINminmax( in_iX, 0, Width() - 1 ),
                   WINminmax( in_iY, 0, Height() - 1 ) );
}
/// Does min/max
inline float IMAGERaw::Safe( const int in_iX, const int in_iY, const int in_iRGB ) const
{
    return (*this)( WINminmax( in_iX, 0, Width() - 1 ),
                   WINminmax( in_iY, 0, Height() - 1 ), WINminmax(in_iRGB, 0, 2) );
}


/** Image manipulation.
 
 A 1D image
 
 <br>Files:
 <dd><li> include/utils/Image_Raw1D.H
 <dd><li> src/utils/image/IMAGERaw1D.cpp
 */
class IMAGERaw1D {
protected:
    
    int           m_iXSize, m_iYSize;
    vector<int>    m_aiXIndx;
    vector<float>  m_afImage;
    
public:
    /**@name Accessors */
    //@{
    ///
    int     Width()  const { return m_iXSize; }
    ///
    int     Height() const { return m_iYSize; }
    ///
    int     NPixels() const { return m_iXSize * m_iYSize; }
    
    ///
    float     operator()( const int in_iX, const int in_iY ) const;
    ///
    float operator()( const R2Pt & ) const;
    ///
    float operator()( const int in_iPixel ) const { return m_afImage[in_iPixel]; }
    ///
    int        Index     ( const int in_iX, const int in_iY ) const;
    ///
    int        X         ( const int in_iIndex ) const;
    ///
    int        Y         ( const int in_iIndex ) const;
    ///
    float Interpolate( const R2Pt &in_pt ) const;
    ///
    vector<R2Pt> InterpolateIndex( const R2Pt &in_pt ) const;
    ///
    const float * Data() const { return &m_afImage[0]; }
    ///
    float * Data() { return &m_afImage[0]; }
    
    ///
    bool    IndexOk( const int in_iX, const int in_iY ) const;
    /// Does min/max
    float Safe( const int in_iX, const int in_iY ) const;
    
    //@}
    
    /**@name Assignements */
    //@{
    ///
    void   SetSize( const int in_iX, const int in_iY );
    
    ///
    float &operator()( const int in_iX, const int in_iY ) ;
    ///
    void    Set       ( const int in_iPixel, const float &in_pix );
    ///
    void    Set       ( const int in_iX, const int in_iY, const float &in_pix );
    ///
    void    Set       ( const float &in_pix );
    
    //@}
    
    /**@name Resampling */
    //@{
    ///
    IMAGERaw1D &operator=( const IMAGERaw1D & ) ;
    //@}
    
    ///
    IMAGERaw1D();
    ///
    IMAGERaw1D( const int in_iX, const int in_iY);
    ///
    ~IMAGERaw1D();
    
    /**@name Read/write */
    //@{
    /// Format determined by filename extension
    bool Read( const char *in_strFName );
    /// Format determined by filename extension
    bool Write( const char *in_strFName ) const;
    //@}
};

inline int IMAGERaw1D::Index( const int in_iX, const int in_iY ) const
{
    ASSERT( in_iX >=0 && in_iX < Width() );
    return m_aiXIndx[ in_iY ] + in_iX;
}

inline float  IMAGERaw1D::operator()( const int in_iX, const int in_iY) const
{
    return m_afImage[ Index(in_iX, in_iY) ];
}

inline float &IMAGERaw1D::operator()( const int in_iX, const int in_iY) 
{
    return m_afImage[ Index(in_iX, in_iY) ];
}

inline float IMAGERaw1D::operator()( const R2Pt &in_pt ) const
{
    const int iIndx =  Index( WINminmax( (int) (in_pt[0] * m_iXSize + 0.5), 0, m_iXSize - 1 ),
                             WINminmax( (int) (in_pt[1] * m_iYSize + 0.5), 0, m_iYSize - 1 ) );
    
    return m_afImage[iIndx];
}


inline void IMAGERaw1D::Set( const int in_iX, const int in_iY, const float &in_pix ) 
{
    const int iIndx = Index(in_iX, in_iY);
    
    m_afImage[ iIndx ] = in_pix;
}

inline void IMAGERaw1D::Set( const int in_iIndx, const float &in_pix ) 
{
    m_afImage[ in_iIndx ] = in_pix;
}


inline int IMAGERaw1D::X( const int in_iIndex ) const
{
    return (in_iIndex) % Width();
}

inline int IMAGERaw1D::Y( const int in_iIndex ) const
{
    return (in_iIndex) / Width();
}

inline bool IMAGERaw1D::IndexOk( const int in_iX, const int in_iY ) const
{
    if ( in_iX < 0 || in_iY < 0 )
        return false;
    
    if ( in_iX >= Width() || in_iY >= Height() )
        return false;
    
    return true;
}

inline float IMAGERaw1D::Safe( const int in_iX, const int in_iY ) const
{
    return (*this)( WINminmax( in_iX, 0, Width() - 1 ),
                   WINminmax( in_iY, 0, Height() - 1 ) );
}

extern void OGLImages_GetImage( IMAGERaw & );

#endif

