/*
 *  OGLRoutines.cpp
 *  Painter3D
 *
 *  Created by Cindy Grimm on 2/27/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#include "SystemSpecific.h"
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include "OGLRoutines.h"

#define LOC() __FILE__, __LINE__

CGcontext *contextSingleton = NULL;

void ShaderError(char *pFile, int pLine, char *pText, ...)
{
    char error[2048];
    va_list valist;
    va_start(valist, pText);
    vsprintf(error, pText, valist);
    va_end(valist);
    fprintf(stderr, "ERROR: %s(%i): %s\n", pFile, pLine, error);
    const char *listing = cgGetLastListing(*contextSingleton);
    if (listing)
        fprintf(stderr, "%s\n", listing);
} 

void ShaderWarning(char *pFile, int pLine, char *pText, ...)
{
    char error[2048];
    va_list valist;
    va_start(valist, pText);
    vsprintf(error, pText, valist);
    va_end(valist);
    fprintf(stderr, "Warning : %s(%i): %s\n", pFile, pLine, error);
  	const char *listing = cgGetLastListing(*contextSingleton);
    if (listing)
		fprintf(stderr, "%s\n", listing);
} 

void CgErrorCallback() {
	fprintf(stderr, "\n**Cg Error:\n-----------\n%s\n", cgGetErrorString(cgGetError()));
	const char *listing = cgGetLastListing(*contextSingleton);
    if (listing)
		fprintf(stderr, "%s\n", listing);
	fprintf(stderr, "-----------\n\n");
	fflush(stderr);
}

const char * OGLShader2::GetTechnique(int pIndex) {
    if(namesList.size() == 0) {
        return NULL;
    }
    return namesList[pIndex];
};

OGLShader2::OGLShader2() {
	context = cgCreateContext();
	contextSingleton = &context;
    
	cgSetErrorCallback(CgErrorCallback);
	cgGLRegisterStates(context);
    
	numTechniques = 0;
	loaded = false;
	complete = false;
	firstPass = true;
}

bool OGLShader2::Validate() {
	bool validTech = false;
	technique = cgGetFirstTechnique(effect);
    while (technique) {
		if (cgValidateTechnique(technique) == CG_FALSE) {
			ShaderWarning(LOC(), "Technique %s did not validate.  Skipping.\n", cgGetTechniqueName(technique));
		}
		else {
			fprintf(stderr, "Valid technique(%i): %s.\n", numTechniques, cgGetTechniqueName(technique));
			numTechniques++;
			validTech = true;
			techniqueList.push_back(technique);
			namesList.push_back(cgGetTechniqueName(technique));
		}
        technique = cgGetNextTechnique(technique);
    }
    
	if(!validTech) {
		ShaderError(LOC(), "No valid techniques");
		return false;
	}
    
	CGparameter p = cgGetFirstEffectParameter(effect);
	while(p) {
		printf("'%s':\nBase:\t\t%s\nType:\t\t%s\nSemantic:\t%s\nArraySize:\t%i\nRows:\t\t%i\nCols:\t\t%i\n------\n", cgGetParameterName(p), cgGetTypeString(cgGetParameterBaseType(p)), cgGetTypeString(cgGetParameterType(p)), cgGetParameterSemantic(p), cgGetArraySize(p, 0), cgGetParameterRows(p), cgGetParameterColumns(p));
		p = cgGetNextParameter(p);
	}
	loaded = true;
	return true;
}

bool OGLShader2::LoadShaderFromFile(const char * pFileName) {
    
	fprintf(stderr, "Creating shader file");
    
	if(loaded) {
		ShaderError(LOC(), "Already loaded shader, please create new OGLShader object");
		return false;
	}
    
	if(pFileName == NULL) {
		ShaderError(LOC(), "Bad Shader File");
		return false;
	}
    
	effect = cgCreateEffectFromFile(context, pFileName, NULL);
	if(!effect) {
		ShaderError(LOC(), "Couldn't create effect file: '%s'", pFileName);
		return false;
	}
    
	Validate();
	loaded = true;
	return true;
}

bool OGLShader2::SetTechnique(char *pTech) {
	if(!loaded) {
		ShaderError(LOC(), "Shader not loaded!");
		return false;
	}
    
	technique = cgGetFirstTechnique(effect);
	while(technique && strcmp(pTech, cgGetTechniqueName(technique))) {
		technique = cgGetNextTechnique(technique);
	}
    
	//Technique set
	pass = cgGetFirstPass(technique);
    
	return true;
}


bool OGLShader2::StartPass() {
	if(!loaded) {
		ShaderError(LOC(), "Shader not loaded!");
		return false;
	}
	if(firstPass) {
		pass = cgGetFirstPass(technique);
	}
	cgSetPassState(pass);
	return true;
}

bool OGLShader2::EndPass() {
	if(!loaded) {
		ShaderError(LOC(), "Shader not loaded!");
		return false;
	}
	cgResetPassState(pass);
	pass = cgGetNextPass(pass);
	if(!pass) {
		complete = true;
	}
	return true;
}

// Necessary to flip!
bool OGLShader2::SetOGLMatrix4x4BySemantic(char * pSemantic, const R4Matrix &pMat) {
	CGparameter p = cgGetFirstEffectParameter(effect);
	while(p) {
		p = cgGetNextParameter(p);
	}
    
	CGparameter m = cgGetEffectParameterBySemantic(effect, pSemantic);
	cgSetMatrixParameterdr(m, &pMat(0,0));
	return true;
}

bool OGLShader2::SetBoolBySemantic(char * pSemantic, bool pValue) {
	return SetFloatBySemantic(pSemantic, pValue);
};

bool OGLShader2::SetFloat3BySemantic(char * pSemantic, float pVec[3]) {
	CGparameter p = cgGetEffectParameterBySemantic(effect, pSemantic);
	cgSetParameter3fv(p, &pVec[0]);
	return true;
};

bool OGLShader2::SetTextureBySemantic(char * pSemantic, GLuint pID) 
{
	bool bFound = false;
	for ( unsigned int i = 0; i < m_aTexIds.size(); i++ ) {
		if ( m_aTexIds[i].first == pID )
			bFound = true;
	}
	if ( bFound == false ) {
		m_aTexIds.push_back( pair<GLuint, CGparameter>( pID, cgGetEffectParameterBySemantic( effect, pSemantic) ) );
	    cgGLSetTextureParameter( m_aTexIds.back().second, pID );
	}
    
	return true;
}


bool OGLShader2::SetFloatBySemantic(char * pSemantic, float pValue) {
	CGparameter p = cgGetEffectParameterBySemantic(effect, pSemantic);
	cgSetParameter1d(p, pValue);
	return true;
};

bool OGLShader2::SetIntBySemantic(char * pSemantic, int pValue) {
	CGparameter p = cgGetEffectParameterBySemantic(effect, pSemantic);
	cgSetParameter1i(p, pValue);
	return true;
};

OGLShader2::~OGLShader2() {
    
}

void OGLBegin2D(  const int in_iWidth, const int in_iHeight )
{
    ASSERT(in_iWidth > 1 && in_iHeight > 1);
    ::glViewport(0, 0, in_iWidth, in_iHeight);
    
    glMatrixMode( GL_PROJECTION );
    glPushMatrix();
    glLoadIdentity();
    
    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();
    glLoadIdentity();
    
    glPushAttrib( GL_LIGHTING );
    glDisable( GL_LIGHTING );
}

void OGLEnd2D( )
{
    glMatrixMode( GL_PROJECTION );
    glPopMatrix();
    
    glMatrixMode( GL_MODELVIEW );
    glPopMatrix();
    
    glPopAttrib( );
}

void OGLDrawCircle( const R2Pt &in_pt, const double in_dR )
{
    if ( fabs( in_dR ) < 1e-32 )
        return;
    
    const double dDiv = WINminmax( ( M_PI / 180.0 ) * in_dR, M_PI / 4.0, M_PI / 10000.0 );
    
    glPushMatrix();
    glTranslated( in_pt[0], in_pt[1], 0 );
    
    glBegin( GL_LINE_LOOP );
    for ( double dTheta = 0.0; dTheta < 2.0 * M_PI; dTheta += dDiv ) {
        glVertex2d( in_dR * cos( dTheta ), in_dR * sin( dTheta ) );
    }
    glEnd();
    
    glPopMatrix();
}

